package cds.stage;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;

import javax.swing.JPanel;

import cds.aladin.Coord;
import cds.tools.pixtools.CDSHealpix;

/**
 * Maquette pour le prototypage sous ANDROID d'un ciel comlet dessin en HEALPix niveau 3 (768 losanges)
 * @author Pierre Fernique [CDS]
 * @version 1.0 cration - avril 2011
 */
public class MySky extends JPanel implements MouseMotionListener, MouseListener,MouseWheelListener {
   
   /** Largeur de la fentre */
   private final static int WIDTH=800;
   
   /** Hauteur de la fentre */
   private final static int HEIGHT=500;
   
   /** Facteur initial du zoom */
   private final static double INITZOOM = 0.03;
   
   /** Order HEALPix 3 => 768 losanges */
   private final static int ORDER = 3;
   
   /** Nombre de losanges HEALPix (pour order=3 => 768) */
   private final static int MAXNPIX = 12*(int)Math.pow(4,ORDER);
   
   /** Dernier point cliqu */
   private double xref=0;       
   private double yref=0;
   
   /** Projection courante */
   private Proj proj;
   
   MySky() {
      proj = new Proj(WIDTH/2,HEIGHT/2,INITZOOM);
      addMouseWheelListener(this);
      addMouseListener(this);
      addMouseMotionListener(this);
      setBackground(Color.white);
   }
   
   /** Dessine le ciel => en premire approximation, tous les bords des
    * losanges HEALPix */
   private void drawSky(Graphics g) {
      for( int npix=0; npix<MAXNPIX; npix++ ) {
         Coord [] corners = computeHpxCorners(ORDER,npix);
         if( corners==null ) continue;
         drawLosange(g,corners);
      }
   }

   /** Dessine un losange */
   private void drawLosange(Graphics g,Coord [] corners) {
      for( int i=0; i<4; i++ ) {
         proj.computeXY(corners[i]);
         if( Double.isNaN(corners[i].x) ) return;
      }
      g.drawLine( (int)corners[0].x,(int)corners[0].y, (int)corners[1].x,(int)corners[1].y);
      g.drawLine( (int)corners[1].x,(int)corners[1].y, (int)corners[3].x,(int)corners[3].y);
      g.drawLine( (int)corners[3].x,(int)corners[3].y, (int)corners[2].x,(int)corners[2].y);
      g.drawLine( (int)corners[2].x,(int)corners[2].y, (int)corners[0].x,(int)corners[0].y);
   }
   
   /** Calcul les 4 coins d'un losange HEALPix en coordonnes clestes */
   private Coord [] computeHpxCorners(int order, long npix) {
      try {
         long nside = (long)Math.pow(2,order);
         double [][] x = CDSHealpix.corners_nest(nside,npix);
         Coord [] corners = new Coord[4];
         for( int i=0; i<x.length; i++ ) corners[i] = new Coord(x[i][0],x[i][1]);
         return corners;
      } catch( Exception e ) { }
      return null;
   }
   
   public Dimension getPreferredSize() { return new Dimension(WIDTH,HEIGHT); }
   
   public void paintComponent(Graphics g) {
      super.paintComponent(g);
      g.setColor(Color.blue);
      drawSky(g);
   }

   /** Action sur la roulette de la souris 
    * => augmentation ou diminution du facteur de zoom */
   public void mouseWheelMoved(MouseWheelEvent e) {
      int sens = e.getWheelRotation();
      double z = proj.getZoom();
      if( sens>0 ) z*=2;
      else z/=2;
      proj.setZoom(z);
      repaint();
   }

   public void mouseClicked(MouseEvent e) {
   }

   /** Clic souris => mmorisation de la position */
   public void mousePressed(MouseEvent e) {
      xref = e.getX();
      yref = e.getY();
   }

   public void mouseReleased(MouseEvent e) {
      System.out.println("MySky mouse released");
   }

   public void mouseEntered(MouseEvent e) {
      System.out.println("MySky entered");
   }

   public void mouseExited(MouseEvent e) {
      System.out.println("MySky exited");
   }

   /** Drag souris => changement du centre de projection en fonction
    * de la distance du dplacement (mthode trs simplifie) */
   public void mouseDragged(MouseEvent e) {
      int x = e.getX();
      int y = e.getY();
      double deltaX = x - xref;
      double deltaY = y - yref;
      xref = x;
      yref = y;
      proj.deltaProjCenter(deltaX/2,deltaY/2);
      repaint();
   }

   public void mouseMoved(MouseEvent e) {
   }

}
