// Copyright 2010 - UDS/CNRS
// The Aladin program is distributed under the terms
// of the GNU General Public License version 3.
//
//This file is part of Aladin.
//
//    Aladin is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, version 3 of the License.
//
//    Aladin is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    The GNU General Public License is available in COPYING file
//    along with Aladin.
//

package cds.tools.pixtools;

import healpix.core.HealpixIndex;
import healpix.core.base.set.LongRangeSet;
import healpix.tools.SpatialVector;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/** Wrapper Healpix CDS pour pouvoir passer facilement d'une librairie Healpix  une autre */
public final class CDSHealpix {
   
   static final String [] MODE = {
      "CDSBestof (no debug)",
      "Will'O Mulan [GAIA] (debug mode)",
      "Kuropatkin (JPL) + CDS",
      "Kotek (based on Kuropatkin)"
   
   };
   
   static public final int BESTOF     = 0;
   static public final int WILL       = 1;
   static public final int KUROPATKIN = 2;
   static public final int KOTEK      = 3;
   
   static public final int HEALPIXMAXORDER[] = { 28, 28, 20, 20 };
   
   static public final int MAXMODE=3;
   
   static final String MODEEXCEPTION = "This method is not supported in the current HEALPix library";

   static private int mode = BESTOF;
   
   /** Positionne la librairie CDSHealpix  utiliser */
   static public void setMode(int mode) {
      if( mode==CDSHealpix.mode ) return;
      CDSHealpix.mode=mode;
      System.out.println("Current HEALPix: "+MODE[mode]+" library");
   }
   
   /** Retourne le mode CDSHealpix courant */
   static public int getMode() { return mode; }
   
   /** Pour du debug - changement cyclique de mode */
   static public String switchMode() {
      mode++;
      if( mode>MAXMODE ) mode=0;
      return MODE[mode];
   }
   
   /** Retourne l'ordre maximum support par la librairie courante */
   static public int getMaxOrder() { return HEALPIXMAXORDER[mode]; }

   /** Voir Healpix documentation */
   static public double[] pix2ang_nest(long nside,long ipix) throws Exception {
      switch(mode) {
         case BESTOF:
//         case WILL: synchronized(lockWill) { initWillMode(nside); return hWill.pix2ang_nest(ipix); }
         case WILL: return hWill[ initWillMode(nside) ].pix2ang_nest(ipix);
         default: throw new Exception(MODEEXCEPTION);
      }
   }
   
   /** Voir Healpix documentation */
   static public double[] pix2ang_ring(long nside,long ipix) throws Exception {
      switch(mode) {
         case BESTOF:
         case WILL: return hWill[ initWillMode(nside) ].pix2ang_ring(ipix);
         default: throw new Exception(MODEEXCEPTION);
      }
   }
   
   /** Voir Healpix documentation */
   static public long ang2pix_nest(long nside,double theta, double phi) throws Exception {
      switch(mode) {
         case BESTOF:
//         case WILL: synchronized( lockWill ) { initWillMode(nside); return hWill.ang2pix_nest(theta, phi); }
         case WILL: return hWill[ initWillMode(nside) ].ang2pix_nest(theta, phi);
         default: throw new Exception(MODEEXCEPTION);
      }
   }
   
   /** Voir Healpix documentation */
   static public long ang2pix_ring(long nside,double theta, double phi) throws Exception {
      switch(mode) {
         case BESTOF:
//         case WILL: synchronized( lockWill ) { initWillMode(nside); return hWill.ang2pix_ring(theta, phi); }
         case WILL: return hWill[ initWillMode(nside) ].ang2pix_ring(theta, phi);
         default: throw new Exception(MODEEXCEPTION);
      }
   }
   
   /** Voir Healpix documentation */
   static public long[] query_disc(long nside,double ra, double dec, double radius) throws Exception {
	   return query_disc(nside, ra, dec, radius, true);
   }
   static public long[] query_disc(long nside,double ra, double dec, double radius, boolean inclusive) throws Exception {
	      switch(mode) {
	      case BESTOF:
	         case WILL: return query_discWill(nside,ra,dec,radius, inclusive);
	         default: throw new Exception(MODEEXCEPTION);
	      }	   
   }
   /** Voir Healpix documentation */
   static public long[] query_polygon(long nside,ArrayList<double[]>list) throws Exception {
      switch(mode) {
         case BESTOF:
         case WILL: return query_polygonWill(nside,list);
         default: throw new Exception(MODEEXCEPTION);
      }
   }
   
   /** Voir Healpix documentation */
   static public double[][] corners_nest(long nside,long npix) throws Exception {
      switch(mode) {
         case WILL: return corners_nestWill(nside, npix);
         case BESTOF:
         default: return corners_nestCDS(nside, npix);
      }
   }
   
   /** Voir Healpix documentation */
   static public List<Long> neighbours_nest(long nside, long npix) throws Exception  {
      switch(mode) {
         default: return hWill[ initWillMode(nside) ].neighbours_nest(npix);
      }
   }
   
   /** Voir Healpix documentation */
   static public long nest2ring(long nside, long npix) throws Exception  {
      switch(mode) {
         case BESTOF:
//         case WILL: synchronized( lockWill ) { initWillMode(nside); return hWill.nest2ring(npix); }
         case WILL: return hWill[ initWillMode(nside) ].nest2ring(npix);
         default: throw new Exception(MODEEXCEPTION);
      }
   }
   
   /** Voir Healpix documentation */
   static public double pixRes(long nside) {
      double res = 0.;
      double degrad = Math.toDegrees(1.0);
      double skyArea = 4.*Math.PI*degrad*degrad;
      double arcSecArea = skyArea*3600.*3600.;
      long npixels = 12*nside*nside;
      res = arcSecArea/npixels;
      res = Math.sqrt(res);
      return res;
   }
   
   /** Voir Healpix documentation */
   static public double[] radecToPolar(double[] radec) { return radecToPolar(radec,new double[2]); }
   static public double[] radecToPolar(double[] radec,double polar[]) {
      polar[0] = Math.PI/2. - radec[1]/180.*Math.PI;
      polar[1] = radec[0]/180.*Math.PI;
      return polar;
   }
   
   /** Voir Healpix documentation */
   static public double[] polarToRadec(double[] polar) { return polarToRadec(polar,new double[2]); }
   static public double[] polarToRadec(double[] polar,double radec[]) {
      radec[1] = (Math.PI/2. - polar[0])*180./Math.PI;
      radec[0] = polar[1]*180./Math.PI;
      return radec;
   }
   
   static final private double LOG2 = Math.log(2);
   static public long log2(long x) { return (long)(Math.log(x)/LOG2); }
   
   
   
   // ------------------------------- Particularits Will

//   static private long nsideWill = -1;  // NSIDE courant dans le mode WILL uniquement
   static private HealpixIndex hWill[] = new HealpixIndex[30];  // Objet Healpix dans le mode WILL uniquement
//   static private Object lockWill = new Object();
   
   // Initialisation de l'objet de manipulation Healpix (dans le mode WILL uniquement)
//   static private void initWillMode(long nside) throws Exception {
//      if( nsideWill==nside ) return;
//      hWill = new HealpixIndex((int)nside);
//      nsideWill=nside;
//   }
   static private int initWillMode(long nside) throws Exception {
      if( hWill==null ) hWill = new HealpixIndex[30];
      int order = (int)log2(nside);
      if( hWill[order]!=null ) return order;
      hWill[order] = new HealpixIndex((int)nside);
//      Aladin.aladin.trace(4,"CDSHealpix.initWillMode(2^"+order+")");
      return order;
   }
   
   static private long[] query_discWill(long nside,double ra, double dec, double radius, boolean inclusive) throws Exception {
//      synchronized( lockWill ) {
         SpatialVector vector = new SpatialVector(ra,dec);
         int order = initWillMode(nside);
//         LongRangeSet list = hWill.queryDisc(vector, radius, 1, inclusive?1:0);
         LongRangeSet list = hWill[order].queryDisc(vector, radius, 1, inclusive?1:0);
         if( list==null ) return new long[0];
         return list.toArray();
//      }
   }

   static private long[] query_polygonWill(long nside,ArrayList<double[]> cooList) throws Exception {
//      synchronized( lockWill ) {
         ArrayList vlist = new ArrayList(cooList.size());
         Iterator<double[]> it = cooList.iterator();
         while( it.hasNext() ) {
            double coo[] = it.next();
            vlist.add(new SpatialVector(coo[0], coo[1]));

         }
         int order = initWillMode(nside);
//         LongRangeSet list = hWill.query_polygon((int)nside, vlist, 1, 1);
         LongRangeSet list = hWill[order].query_polygon((int)nside, vlist, 1, 1);
         if( list==null ) return new long[0];
         return list.toArray();
//      }
   }      
   
   static final int [] A = { 3, 2, 0, 1 };
   static public double[][] corners_nestWill(long nside,long npix) throws Exception {
      int order = initWillMode(nside);
      SpatialVector [] v = hWill[order].corners_nest(npix, 1);
      double [][] corners = new double[v.length][2];
      for( int i=0; i<v.length; i++ ) {
         int j = A[i];
         corners[j][0] = v[i].ra();
         corners[j][1] = v[i].dec();
      }
      return corners;  
   }
   
//   public static void main(String[] args) {
//      try {
//         int nside = 8;
//         SpatialVector vector = new SpatialVector(219.92904166666668,85.88719444444445);
//         double radius = 3.91698480573189;
//
//         HealpixIndex hi = new HealpixIndex(nside);
//         LongRangeSet vlist = hi.queryDisc(vector, radius / 180 * Constants.PI, 1, 1);
//         long [] list = vlist==null ? new long[0] : vlist.toArray();
//
//         System.out.print("ra="+vector.ra()+ " dec="+vector.dec()+" radius="+radius+" Nside="+nside+" => npixlist:");
//         for( int i = 0; i < list.length; i++ ) {
//            System.out.print(" " + list[i]);
//         }
//         System.out.println();
//      } catch( Exception e ) {
//         e.printStackTrace();
//      }
//   }


   // ------------------------------- Particularits Kuropatkin + CDS
   
   
   /** Approximation des coordonnes RA,DEC des 4 angles du losange
    * Je recherche les coord (centrales) des losanges des 4 coins dans la rsolution
    * Healpix maximale */
   private static double [][] corners_nestCDS(long nside, long pixid) {
      double [][] corners;
      try {
         int order = (int)log2(nside);
         corners = new double[4][2];
         int orderFile = getMaxOrder() - order; // Je ne dois pas dpasser la limite Healpix
         long nSidePix = (long) Math.pow(2, orderFile);
         // Numro des pixels des 4 coins
         long c0, c1, c2, c3;
         c0 = c1 = c2 = 0;
         c3 = (nSidePix * nSidePix) - 1;
         for( int i = 0; i < orderFile; i++ ) c1 = (c1 << 2) | 1;
         for( int i = 0; i < orderFile; i++ ) c2 = (c2 << 2) | 2;
         // Chaque pixel "interne" va tre remplac par nsidePix*nsidePix pixels
         // d'o l'offset suivant
         long offset = pixid * nSidePix * nSidePix;
         c0 += offset;
         c1 += offset;
         c2 += offset;
         c3 += offset;
         long nSideFile = (long) Math.pow(2, order + orderFile);
         polarToRadec(pix2ang_nest(nSideFile, c0) , corners[0]);
         polarToRadec(pix2ang_nest(nSideFile, c1) , corners[1]);
         polarToRadec(pix2ang_nest(nSideFile, c2) , corners[2]);
         polarToRadec(pix2ang_nest(nSideFile, c3) , corners[3]);
       } catch( Exception e ) { return null; }
      return corners;
   }

   


}
